/*
 * TAB 4 ------------------------------------------------------------ lline
 * ('long line') - locate & display the longest lines in a text file ray
 * berry - bellevue wa CIS 73407,3152 uucp: ...!ole!ray
 *
 * This program is a derivative of (and was directly inspired by) a program
 * called 'longest.c' by Larry Gensch.  Approximately 30% of the code in this
 * program is directly attributable to him. His original work may be found in
 * the unixforum sig of CIS.
 *
 * the following modifications are made in this version:
 *
 * - deleted 'w' option (width option for output fields) - deleted 'c' option
 * (set char separating output fields) - converted 'getlen' function to 8086
 * ASM for speed - added support for handling backspaces - added 'p' option
 * to print longest line found - added 'r' option to enumerate cnts of
 * longest line lengths - changed numerous types to optimize program speed
 * for pc - other minor changes in src specific to pc compilers + v 1.1
 * 9/13/89 rjb - added report for trailing whitespace - added report of tab
 * characters in file + v 2.0 9/25/89 rjb - added FILTER output option with
 * tab expand, white strip, optional indent, backspace stripping
 *
 * limitations of this version: max 65535 lines/file, 255 chars/line
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifndef HEADER
# define	HEADER_DEFAULT	0
#else
# define	HEADER_DEFAULT	1
#endif

#define	VERSION	"lline ('longest line') 2.0 9/25/89 Ray Berry, "\
		"Bellevue WA. - CIS 73407,3152"

#define RCNT 6			/* number of lengths reported for 'r' option */
#define STR(x) #x
#define QSTR(x) STR(x)
#define	MAX_LINE 256		/* max length of a line */
#define MAX_INDENT 40		/* max number of spaces for indent */

FILE           *ofile = stdout;

unsigned        linecnt[MAX_LINE];	/* used to totalize line lengths */
char            sflag,		/* show mode - show 1st instance longest line */
                rflag,		/* report - print summary cnt of longest
				 * lines */
                wflag,		/* whitespace - print each line w/trailing
				 * whitespace */
                bflag,		/* backspace - strip backspace (-f only) */
                xflag;		/* truncate */
unsigned        has_tabs;
char            trunc_size, tab_size = 8;
unsigned        white_max;	/* flag denotes presence of trlng whtsp in
				 * longest ln */
char            inbuf[MAX_LINE], outbuf[MAX_LINE];	/* used in analyze mode */
int             indent_size;

char           *usage1 =
"usage:  lline [-h] [-o file] [-r] [-f] [-t#] [-w] [-V] [file...(*? ok)]\n"
"ANALYZE (according to following flags) or FILTER (below) text files\n";
char           *usage2 =
#ifdef HEADER
"    -h         Do not display header line.\n"
#else
"    -h         Display header line (default no header)\n"
#endif
"    -r         Report # of occurences of "
QSTR(RCNT) " longest line lengths\n"
"    -s         Show 1st instance of line of longest length\n"
"    -t #       Set (expand for -f) tab spacing to # (default 8)\n"
"    -w         Report trailing whitespace as encountered\n"
"    -V         Print program version information & exit\n"
"    -o file    Send output to 'file' (default stdout)\n";
char           *usage3 =
"    -f         FILTER mode- copy stdin>stdout according to -t,-o (above) and:\n"
"         -i #       Indent each output line by # spaces\n"
"         -x #       truncate (indented) line length to max # chars\n"
"         -b         strip all backspaces from the output\n"
"               Filter mode expands tabs & strips trailing whitespace.\n"
"    If no files specified on command line, stdin is read\n"
"    Specifying a file name of '-' specifies stdin explicitly.\n";

unsigned pascal getlen();
int             chkfile(), printfile();
int             (*EACH_FILE) () = chkfile;

/*
 * main program -- collect options and longest specified files (or stdin):
 */
char           *basename();
main(int argc, char *argv[])
{
	int             header = HEADER_DEFAULT;
	extern char    *optarg;
	extern int      optind;
	int             option;

	argv[0] = basename(argv[0]);
	while ((option = getopt(argc, argv, "?HhXx:RrTt:Oo:FfSsWwVvIi:Bb")) != EOF) {
		/*
		 * note - my getopt is hacked to recognize either '-' or '/'
		 * as an option leadin  - and returns '!' for unrecognized
		 * options - rjb
		 */
		switch (tolower(option)) {
		case 'h':
			header ^= 1;
			break;
		case 'o':
			ofile = fopen(optarg, "w");
			if (ofile == NULL)
				fatal(optarg);
			break;
		case 'f':
			EACH_FILE = printfile;	/* EACHFILE = function ptr */
			break;
		case 'b':
			bflag++;
			break;
		case 'r':
			rflag++;
			break;
		case 's':
			sflag++;
			break;
		case 'w':
			wflag++;
			break;
		case 'x':
			xflag++;
			trunc_size = atoi(optarg);
			if (trunc_size < 1 || tab_size > MAX_LINE-1) {
				fprintf(stderr, "illegal truncate argument\n");
				exit(1);
			}
			break;
		case 't':
			tab_size = atoi(optarg);
			if (tab_size < 2 || tab_size > 32) {
				fprintf(stderr, "tab must be between 2 and 32\n");
				exit(1);
			}
			break;
		case 'i':
			indent_size = atoi(optarg);
			if (indent_size < 1 || indent_size > MAX_INDENT-1) {
				fprintf(stderr, "indent must be in range 1-"
					QSTR(MAX_INDENT) "\n");
				exit(1);
			}
			break;
		case 'v':
			(void) fprintf(stderr, "%s\n\n", VERSION);
			exit(0);
		case '?':
		case '!':
		default:
			fputs(usage1, stderr);
			fputs(usage2, stderr);
			fputs(usage3, stderr);
			exit(option != 'V');
			}
	}

	if (header && EACH_FILE != printfile)
		fputs(" length    line file\n", ofile);

	if (optind == argc)
		EACH_FILE(stdin, "<stdin>");
	else
		for (; optind < argc; optind++) {
			memset(linecnt, 0, MAX_LINE - 1);
			if (strcmp(argv[optind], "-") == 0)
				EACH_FILE(stdin, "<stdin>");
			else {
				FILE           *fp = fopen(argv[optind], "r");
				(void) strlwr(argv[optind]);
				if (fp == NULL)
					fatal(argv[optind]);
				EACH_FILE(fp, argv[optind]);
				if (fclose(fp))
					fatal(argv[optind]);
			}
		}
	exit(0);
	/* NOTREACHED */
}

/* printfile - print mode processor for outputting modified text */
printfile(FILE * fp, char *file)
{
	int             ipos, j;
	char           *ip, *op;

	while (fgets(inbuf, MAX_LINE, fp) != NULL) {
		ip = inbuf;
		op = outbuf + indent_size;
		ipos = 0;
		memset(outbuf, ' ', MAX_LINE - 1);
		do {
			switch (*ip) {
			case '\t':
				do
					op++, ipos++;
				while (ipos % tab_size);
				break;
			case '\b':
				if (bflag) {
					ipos--;
					op--;
					break;
				}
			default:
				*op++ = *ip;
				ipos++;
			}
			ip++;
		} while (*(ip - 1));
		*op = '\0';
		j = strlen(outbuf) - 2;	/* strip trailing whitespace */
		while (isspace(outbuf[j--]));
		outbuf[j + 2] = '\n';
		outbuf[j + 3] = '\0';
		if (xflag) {
			outbuf[trunc_size] = '\n';
			outbuf[trunc_size + 1] = '\0';
		}
		fputs(outbuf, ofile);
	}
}
/*
 * chkfile() - analyze lines for length, whitespace, etc.
 */

chkfile(FILE * fp, char *filep)
{
	int             i, j;
	unsigned        len = 0, lineno = 1, white_lines = 0,	/* # of lines w/
								 * trailing white */
	                max = 0, maxline = 0;
	char           *p, filen[32];

	has_tabs = 0;
	strcpy(filen, filep);
	strcat(filen, ":");
	for (; fgets(inbuf, MAX_LINE, fp) != NULL; lineno++) {
		len = strlen(inbuf);
		inbuf[--len] = 0;	/* discard newline */
		/* check for trailing whitespace */
		for (j = len, i = 0; isspace(inbuf[j - 1]); i++, j--);
		if (i) {
			white_lines++;
			if (wflag)
				fprintf(ofile, "%s line %d: %d trailing whitespace char(s)\n", filen, lineno, i);
		}
		linecnt[(len = getlen(inbuf))]++;
		if (len > max) {
			maxline = lineno;
			max = len;
			if (sflag)
				strcpy(outbuf, inbuf);
			white_max = i;	/* trail. white in longest line? */
		}
	}
	fprintf(ofile, "%7u %7u %-13s", max, maxline, filen);
	if (has_tabs || white_lines) {
		if (white_lines)
			fprintf(ofile, "trailing whitespace on %d line(s)", white_lines);
		if (white_lines && has_tabs)
			fprintf(ofile, ", ");
		if (has_tabs)
			fprintf(ofile, "has Tabs");
	}
	fputc('\n', ofile);
	if (sflag)
		puts(outbuf);
	if (rflag) {
		i = MAX_LINE;
		while (!linecnt[--i]);
		for (j = RCNT; j-- && i > 0;) {
			fprintf(ofile, "%4u: %-6u", i, linecnt[i]);
			while (!linecnt[--i]);
		}
		putchar('\n');
	}
	if (white_max)
		fprintf(ofile, "%29sNOTE: longest line has trailing whitespace\n", "");
}

/*
 * fatal() - Display an error message and abort the program.
 */

fatal(char *msg)
{
	perror(msg);
	exit(3);
	/* NOTREACHED */
}
#ifdef MSDOS
_setenvp()
{
}
#endif
